from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import numpy as np
from pydantic import validator
from typing_extensions import Self

from qcio.constants import BOHR_TO_ANGSTROM
from qcio.helper_types import ArrayLike2D

from .base_model import QCIOModelBase

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs

__all__ = ["Molecule", "Identifiers"]


class Identifiers(QCIOModelBase):
    """Molecule identifiers.

    Attributes:
        name_IUPAC: The IUPAC name of the molecule.
        name_common: the common name of the molecule.
        molecule_hash: A hash of the molecule.
        smiles: The SMILES representation of the molecule.
        inchi: The InChI representation of the molecule.
        inchikey: The InChIKey representation of the molecule.
        canonical_explicit_hydrogen_smiles: The canonical explicit hydrogen SMILES
            representation of the molecule.
        canonical_isomeric_explicit_hydrogen_mapped_smiles: The canonical isomeric
            explicit hydrogen mapped SMILES representation of the molecule.
        canonical_isomeric_explicit_hydrogen_smiles: The canonical isomeric explicit
            hydrogen SMILES representation of the molecule.
        canonical_isomeric_smiles: The canonical isomeric SMILES representation of the
            molecule.
        canonical_smiles: The canonical SMILES representation of the molecule.
        pubchem_cid: The PubChem Compound ID of the molecule.
        pubchem_sid: The PubChem Substance ID of the molecule.
        pubchem_conformerid: The PubChem Conformer ID of the molecule.
    """

    name_IUPAC: Optional[str] = None
    name_common: Optional[str] = None
    molecule_hash: Optional[str] = None
    smiles: Optional[str] = None
    inchi: Optional[str] = None
    inchikey: Optional[str] = None
    canonical_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_mapped_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_smiles: Optional[str] = None
    canonical_smiles: Optional[str] = None
    pubchem_cid: Optional[str] = None
    pubchem_sid: Optional[str] = None
    pubchem_conformerid: Optional[str] = None


# class Bond(QCIOModelBase):
#     """A bond object.

#     Attributes:
#         indices: The indices of the atoms in the bond.
#         order: The order of the bond.
#     """

#     indices: Tuple[int, int]
#     order: Optional[float]


class Molecule(QCIOModelBase):
    """A molecule object.

    Attributes:
        symbols: The atomic symbols of the molecule.
        geometry: The geometry of the molecule in Cartesian coordinates. Units are (AU)
            Bohr.
        identifiers: Identifiers for the molecule such as name, smiles, etc.
        charge: The molecular charge.
        connectivity: Explicit description of the bonds between atoms.
        multiplicity: The molecular multiplicity.
        masses: Explicit masses for the atoms. If not set, default isotopic masses are
            used.
    """

    symbols: List[str]
    geometry: ArrayLike2D
    charge: int = 0
    multiplicity: int = 1
    identifiers: Identifiers = Identifiers()
    # masses: Optional[List[float]] = None
    # connectivity: Optional[List[Bond]] = None

    def __repr_args__(self) -> "ReprArgs":
        """A helper for __repr__ that returns a list of tuples of the form
        (name, value).
        """
        return [("name", self.identifiers.name_common), ("symbols", self.symbols)]

    @validator("geometry")
    def shape_n_by_3(cls, v, values, **kwargs):
        """Ensure there is an x, y, and z coordinate for each atom."""
        n_atoms = len(values["symbols"])
        return np.array(v).reshape(n_atoms, 3)

    @classmethod
    def open(cls, filepath: Union[Path, str]) -> Self:
        """Open a molecule from a file.

        Args:
            filepath: The path to the file to open. Maybe a path to a JSON file or an
                XYZ file.
        """

        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            return cls._from_xyz(filepath)
        return super().open(filepath)

    def save(
        self, filepath: Union[Path, str], exclude_none: bool = True, **kwargs
    ) -> None:
        """Save a molecule to a file.

        Args:
            filepath: The path to save the molecule to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Notes:
            If the filepath has a .xyz extension, the molecule will be saved to an XYZ
            file.
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            self._to_xyz(filepath)
            return
        super().save(filepath, **kwargs)

    def _to_xyz(self, filepath: Union[Path, str]) -> None:
        """Write a molecule to an XYZ file.

        Notes:
            Will export qcio data to the comments line with a qcio_key=value format.
        """
        filepath = Path(filepath)

        qcio_kwargs = {
            "qcio_charge": self.charge,
            "qcio_multiplicity": self.multiplicity,
        }
        assert isinstance(self.geometry, np.ndarray)  # For mypy
        geometry_angstrom = self.geometry * BOHR_TO_ANGSTROM

        with open(filepath, "w") as f:
            f.write(f"{len(self.symbols)}\n")
            f.write(f"{' '.join([f'{k}={v}' for k, v in qcio_kwargs.items()])}\n")
            for symbol, (x, y, z) in zip(self.symbols, geometry_angstrom):
                f.write(f"{symbol:2s} {x: >18.12f} {y: >18.12f} {z: >18.12f}\n")

    @classmethod
    def _from_xyz(cls, filepath: Union[Path, str]) -> Self:
        """Create a molecule from an XYZ file.

        Notes:
            Will read qcio data from the comments line with a qcio_key=value format.
        """
        filepath = Path(filepath)

        with open(filepath, "r") as f:
            lines = f.readlines()

        num_atoms = int(lines[0])
        qcio_kwargs: Dict[str, Any] = {
            item.split("=")[0].replace("qcio_", ""): item.split("=")[1]
            for item in lines[1].strip().split()
            if item.startswith("qcio_")
        }
        symbols = []
        geometry = []
        for line in lines[2 : 2 + num_atoms]:
            split_line = line.split()
            symbols.append(split_line[0])
            geometry.append([float(val) / BOHR_TO_ANGSTROM for val in split_line[1:]])

        return cls(symbols=symbols, geometry=geometry, **qcio_kwargs)



from .models.base_io import *  # noqa: F403
from .models.file import *  # noqa: F403
from .models.molecule import *  # noqa: F403
from .models.single_point import *  # noqa: F403

__version__ = "0.1.0"

__all__ = [  # noqa: F405
    # Core Models
    "FileArgs",
    "FileInput",
    "FileSuccessfulOutput",
    "FileFailedOutput",
    "Provenance",
    "Model",
    "SPCalcType",
    "SinglePointProgramArgs",
    "SinglePointInput",
    "SinglePointSuccessfulOutput",
    "SinglePointFailedOutput",
    "SinglePointComputedProperties",
    "Wavefunction",
]



"""File model for handling string and binary data."""

import json
from base64 import b64decode, b64encode
from pathlib import Path
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, validator

from .helper_types import StrOrPath


class Files(BaseModel):
    """Mixin for models that have files that need to be serialized to JSON.

    Attributes:
        files: A dict mapping filename to str or bytes data.
    """

    files: Dict[str, Union[str, bytes]] = {}

    @validator("files", pre=True)
    def convert_base64_to_bytes(cls, value):
        """Convert base64 encoded data to bytes."""
        for filename, data in value.items():
            if isinstance(data, str) and data.startswith("base64:"):
                value[filename] = b64decode(data[7:])
        return value

    def dict(self, *args, **kwargs):
        """Return a dict representation of the object encoding bytes as b64 strings."""
        dict = super().dict(*args, **kwargs)
        files = {}
        for filename, data in self.files.items():
            if isinstance(data, bytes):
                data = f"base64:{b64encode(data).decode('utf-8')}"
            files[filename] = data
        dict["files"] = files
        return dict

    def json(self, *args, **kwargs):
        """Return a JSON representation of the object."""
        return json.dumps(self.dict(*args, **kwargs))

    def add_file(
        self, filepath: Union[Path, str], relative_dir: Optional[Path] = None
    ) -> None:
        """Create a File object from a file on disk.

        Args:
            filepath: The path to the file.
            relative_dir: The directory to make the file relative to. Helpful when
                adding files from a subdirectory.
        """
        filepath = Path(filepath)
        raw_bytes = filepath.read_bytes()
        try:
            data: Union[str, bytes] = raw_bytes.decode("utf-8")  # str
        except UnicodeDecodeError:
            data = raw_bytes  # bytes

        # Set filename relative to relative_dir
        if relative_dir:
            filename = str(filepath.relative_to(relative_dir))
        else:
            filename = filepath.name

        self.files[filename] = data

    def add_files(
        self,
        directory: StrOrPath,
        recursive: bool = False,
        exclude: Optional[List[str]] = None,
    ) -> None:
        """Add all files in a directory to the object.

        Args:
            directory: The directory to add files from.
            recursive: Whether to recursively add files from subdirectories.
            exclude: A list of filenames to exclude from the directory.
        """
        directory = Path(directory)
        if exclude is None:
            exclude = []
        if recursive:
            files = directory.rglob("*")
        else:
            files = directory.glob("*")
        for filepath in files:
            if filepath.is_file() and filepath.name not in exclude:
                self.add_file(filepath, directory)

    def write_files(self, directory: StrOrPath = Path(".")) -> None:
        """Write all files to the specified directory"""
        directory = Path(directory)
        directory.mkdir(exist_ok=True)
        for filename, data in self.files.items():
            mode = "w" if isinstance(data, str) else "wb"
            filepath = directory / filename
            # In case filename is a relative path, create the parent directories
            filepath.parent.mkdir(exist_ok=True, parents=True)
            filepath.open(mode).write(data)



"""Base classes for qcio input and output objects."""
from abc import abstractmethod
from traceback import format_exception
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Type

from ..mixins import Files
from .base_model import QCIOModelBase
from .molecule import Molecule

__all__ = ["Provenance"]


if TYPE_CHECKING:
    from pydantic.typing import ReprArgs


class Provenance(QCIOModelBase):
    """Provenance information for a QC program.

    Attributes:
        program: The name of the program that created the output.
        version: The version of the program that created the output.
        working_dir: The working directory used by the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.

    """

    program: str
    program_version: Optional[str] = None
    working_dir: Optional[str] = None

    def __repr_args__(self) -> "ReprArgs":
        return [("program", self.program), ("program_version", self.program_version)]


class ComputedPropertiesBase(QCIOModelBase):
    """Base class for all properties computed by a program"""

    pass


class ProgramArgsBase(QCIOModelBase, Files):
    """Arguments for a QC program.

    Attributes:
        files: A dict mapping filename to str or bytes data.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """


class StructuredProgramArgsBase(ProgramArgsBase):
    """Base Program Args class for structured qcio input classes.

    This class is used to define the args for a structured qcio input class.

    Attributes:
        keywords: A dict of keywords to be passed to the program. Defaults to an empty
            dict.
        files: Files to be passed to the QC program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    keywords: Dict[str, Any] = {}


class InputBase(QCIOModelBase):
    """Base class for all program input objects.

    Attributes:
        program_args: Input arguments for the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    program_args: ProgramArgsBase

    @abstractmethod
    def get_failed_output_class(self) -> Type["FailedOutputBase"]:
        """Return the FailedOutput class for the input object."""
        raise NotImplementedError()

    @abstractmethod
    def get_successful_output_class(self) -> Type["SuccessfulOutputBase"]:
        """Return the SuccessfulOutput class for the input object."""
        raise NotImplementedError()

    def to_success(
        self,
        provenance: Provenance,
        computed: Optional[ComputedPropertiesBase] = None,
        stdout: Optional[str] = None,
    ) -> "OutputBase":
        """Return a successful output object.

        Args:
            provenance: The provenance information for the output.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        output_class = self.get_successful_output_class()
        return output_class(
            input_data=self, provenance=provenance, computed=computed, stdout=stdout
        )

    def to_failure(self, provenance: Provenance, exc: Exception) -> "FailedOutputBase":
        """Return a failed output object.

        Args:
            provenance: The provenance information for the output.
            exc: The exception that caused the failure.
        """
        output_class = self.get_failed_output_class()
        output_class.update_forward_refs()
        stdout = getattr(exc, "stdout", None)
        computed = getattr(exc, "computed", None)
        traceback = "".join(format_exception(type(exc), exc, exc.__traceback__))

        return output_class(
            input_data=self,
            provenance=provenance,
            traceback=traceback,
            stdout=stdout,
            computed=computed,
        )

    def __repr_args__(self) -> "ReprArgs":
        return [("program_args", self.program_args)]


class StructuredInputBase(InputBase):
    """Base class for all structured qcio input classes.

    Structured inputs consist of args for a QC program and a Molecule.

    Attributes:
        program_args: Structured input arguments for the program.
        molecule: The molecule to be used in the calculation.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.

    """

    program_args: StructuredProgramArgsBase
    molecule: Molecule


class OutputBase(Files, QCIOModelBase):
    """Base class for all output classes.

    Attributes:
        input_data: The input object for the computation.
        success: A boolean indicator that the operation succeeded or failed. Allows
            programmatic assessment of all outputs regardless of if they failed or
            succeeded by checking `output.success`.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
        computed: An object containing the computed properties from the output.
        files: A dict mapping filename to str or bytes of files output by the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    input_data: InputBase
    success: bool
    stdout: Optional[str] = None
    computed: Optional[ComputedPropertiesBase] = None
    provenance: Provenance

    def print_stdout(self) -> None:
        """Print the stdout text"""
        print(self.stdout)


class SuccessfulOutputBase(OutputBase):
    """Base object for any successful computation.

    Attributes:

        input_data: The input object for the computation.
        success: Always True for a successful output.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    success: Literal[True] = True


class FailedOutputBase(OutputBase):
    """Base object for any failed computation.

    Attributes:
        input_data: The input object for the computation.
        success: Always False for a Failed output.
        traceback: String representation of the traceback of the exception that caused
            the failure.
        computed: Any computed data that was able to be extracted before the exception.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    success: Literal[False] = False
    traceback: Optional[str] = None
    computed: Optional[Any] = None

    def __repr_args__(self) -> "ReprArgs":
        """A helper for __repr__ that returns a list of tuples of the form
        (name, value).
        """
        return [
            ("success", self.success),
            ("input_data", self.input_data),
            ("provenance", self.provenance),
        ]



"""Basic file I/O objects for calculations."""
from pathlib import Path
from typing import List, Type, Union

from typing_extensions import Self

from ..mixins import Files
from .base_io import FailedOutputBase, InputBase, ProgramArgsBase, SuccessfulOutputBase

__all__ = ["FileArgs", "FileInput", "FileSuccessfulOutput", "FileFailedOutput"]


class FileArgs(ProgramArgsBase):
    """Program arguments for a FileInput.

    Attributes:
        cmdline_args: A list of command line arguments to be passed to the program.
        files: A dict mapping filename to str or bytes data.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    cmdline_args: List[str] = []


class FileInput(InputBase):
    """File and command line argument inputs for a calculation.

    Attributes:
        program_args: Files and command line arguments for the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    program_args: FileArgs = FileArgs()

    @classmethod
    def from_directory(cls, directory: Union[Path, str], **kwargs) -> Self:
        """Collect all files from directory and instantiate an object."""
        obj = cls(**kwargs)
        directory = Path(directory)
        obj.program_args.add_files(directory)
        return obj

    def get_failed_output_class(self) -> Type["FileFailedOutput"]:
        return FileFailedOutput

    def get_successful_output_class(self) -> Type["FileSuccessfulOutput"]:
        return FileSuccessfulOutput


class FileOutputMixin(Files):
    """Mixin for file based output objects.
    Attributes:
        input_data: The input data for the computation.
        files: A dict mapping filename to str or bytes data.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
        success: A boolean indicator that the operation succeeded (always True)
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
    """

    input_data: FileInput


class FileSuccessfulOutput(SuccessfulOutputBase, FileOutputMixin):
    """Generic output class for pure file based I/O.

    Attributes:
        input_data: The FileInput object used for the computation.
        files: A dict mapping filename to str or bytes data.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
        success: A boolean indicator that the operation succeeded (always True)
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
    """

    input_data: FileInput


class FileFailedOutput(FailedOutputBase, FileOutputMixin):
    """Generic output class for pure file based I/O.

    Attributes:
        input_data: input_data: The FileInput object used for the computation.
        files: A dict mapping filename to str or bytes data.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
        success: Always False for a Failed output.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
    """



"""The Base model from with all QCIO Model objects inherit."""
from abc import ABC
from pathlib import Path
from typing import Any, Dict, Union

import numpy as np
from pydantic import BaseModel, Extra
from typing_extensions import Self


class QCIOModelBase(BaseModel, ABC):
    """Base Model for all QCIO objects.

    Attributes:
        version: The version of the schema.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    # NOTE: No version for now until we have a stable schema
    # or maybe have version be version of code package that generated
    # the object?
    # version: ClassVar[Literal["v1"]] = "v1"
    extras: Dict[str, Any] = {}

    class Config:
        # Raises an error if extra fields are passed to model.
        extra = Extra.forbid
        # Allow numpy types in models. Pydantic will no longer raise an exception for
        # types it doesn't recognize.
        # https://docs.pydantic.dev/latest/usage/types/#arbitrary-types-allowed
        arbitrary_types_allowed = True
        # Don't allow mutation of objects
        # https://docs.pydantic.dev/latest/usage/models/#faux-immutability
        allow_mutation = False
        # convert ndarray to list for JSON serialization
        # https://docs.pydantic.dev/latest/usage/exporting_models/#json_encodershttps://docs.pydantic.dev/latest/usage/exporting_models/#json_encoders  # noqa: E501
        json_encoders = {np.ndarray: lambda v: v.tolist()}
        # exclude fields with value None from serialization
        exclude_none = True

    def dict(self, **kwargs):
        """Convert the object to a dictionary.

        Properly serialize numpy arrays. Serialization is performed in .dict() so that
        multiple string serializers can used it without duplicating logic
        (e.g. json, yaml, toml).
        """
        # Exclude empty lists and dictionaries from serialization
        model_dict = super().dict(**kwargs)
        to_pop = []
        for key, value in model_dict.items():
            if isinstance(value, np.ndarray):
                model_dict[key] = value.tolist()
            elif not value:  # None, empty list, empty dict
                to_pop.append(key)

        for key in to_pop:
            model_dict.pop(key)

        return model_dict

    @classmethod
    def open(cls, filepath: Union[Path, str]) -> Self:
        """Instantiate an object from data saved to disk.

        Args:
            filepath: The path to the object on disk.
            *args: Additional arguments to pass to the constructor.
            **kwargs: Additional keyword arguments to pass to the constructor.

        Returns:
            The instantiated object.
        """
        filepath = Path(filepath)
        return cls.parse_raw(filepath.read_text())
        # pydantic v2
        # return cls.model_validate_json(filepath.read_text())

    def save(
        self, filepath: Union[Path, str], exclude_none: bool = True, **kwargs
    ) -> None:
        """Write an object to disk as json.

        Args:
            filepath: The path to write the object to.
            exclude_none: If True, do not write fields with value None to disk.
            **kwargs: Additional arguments to pass to json().
        """
        filepath = Path(filepath)
        filepath.parent.mkdir(exist_ok=True, parents=True)
        filepath.write_text(self.json(exclude_none=exclude_none, **kwargs))
        # pydantic v2
        # filepath.write(self.model_dump())

    def __repr_args__(self):
        return [
            (name, value)
            for name, value in self.__dict__.items()
            if name != "extras"  # Ignore "extras" field
        ]



from enum import Enum
from typing import TYPE_CHECKING, Optional, Type, Union

import numpy as np
from pydantic import BaseModel, validator

from ..helper_types import ArrayLike2D
from .base_io import (
    ComputedPropertiesBase,
    FailedOutputBase,
    StructuredInputBase,
    StructuredProgramArgsBase,
    SuccessfulOutputBase,
)
from .base_model import QCIOModelBase
from .molecule import Molecule

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs

__all__ = [
    "Model",
    "SPCalcType",
    "SinglePointProgramArgs",
    "SinglePointInput",
    "SinglePointSuccessfulOutput",
    "SinglePointFailedOutput",
    "SinglePointComputedProperties",
    "Wavefunction",
]


class SPCalcType(str, Enum):
    """The single point calculation type."""

    energy = "energy"
    gradient = "gradient"
    hessian = "hessian"

    def __repr__(self) -> str:
        """Custom repr for CalcType"""
        return f"'{self.name}'"


class Model(BaseModel):
    """The model for the quantum chemistry calculation.

    Attributes:
        method: The name of the method to be used in the calculation. Named according to
            the convention of the program being called.
        basis: The name of the basis set to be used in the calculation. Named according
            to the convention of the program being called.
    """

    method: str
    basis: Optional[str] = None


class SinglePointProgramArgs(StructuredProgramArgsBase):
    """Program arguments for a single point calculation.

    Attributes:
        calc_type: The calculation type for the calculation. May be "energy",
            "gradient", or "hessian". Defaults to "energy".
        model: The model for the quantum chemistry calculation.
        keywords: A dict of keywords to be passed to the program. Defaults to an empty
            dict.
        files: Files to be passed to the QC program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    calc_type: SPCalcType
    model: Model

    def __repr_args__(self) -> "ReprArgs":
        """Custom repr for SinglePointProgramArgs"""
        return [
            ("calc_type", self.calc_type),
            ("model", self.model),
        ]


class Wavefunction(QCIOModelBase):
    """The wavefunction for a single point calculation.

    Attributes:
        scf_eigenvalues_a: The SCF alpha-spin orbital eigenvalues.
        scf_eigenvalues_b: The SCF beta-spin orbital eigenvalues.
        scf_occupations_a: The SCF alpha-spin orbital occupations.
        scf_occupations_b: The SCF beta-spin orbital occupations.
    """

    scf_eigenvalues_a: Optional[ArrayLike2D] = None
    scf_eigenvalues_b: Optional[ArrayLike2D] = None
    scf_occupations_a: Optional[ArrayLike2D] = None
    scf_occupations_b: Optional[ArrayLike2D] = None

    _to_numpy = validator(
        "scf_eigenvalues_a",
        "scf_eigenvalues_b",
        "scf_occupations_a",
        "scf_occupations_b",
        allow_reuse=True,
    )(lambda x: np.asarray(x) if x is not None else None)


class SinglePointComputedProperties(ComputedPropertiesBase):
    """The computed properties from a single point calculation.

    Attributes:
        calcinfo_natoms: The number of atoms as computed by the program.
        calcinfo_nalpha: The number of alpha electrons as computed by the program.
        calcinfo_nbeta: The number of beta electrons as computed by the program.
        calcinfo_nbasis: The number of basis functions in the calculation.
        calcinfo_nmo: The number of molecular orbitals in the calculation

        energy: The energy of the molecule in Hartrees.
        gradient: The gradient of the molecule in Hartrees/Bohr.
        hessian: The hessian of the molecule in Hartrees/Bohr^2.
        nuclear_repulsion_energy: The nuclear repulsion energy of the molecule in
            Hartrees.
    """

    # calcinfo contains general information about the calculation
    calcinfo_natoms: Optional[int] = None
    calcinfo_nbasis: Optional[int] = None
    calcinfo_nmo: Optional[int] = None
    calcinfo_nalpha: Optional[int] = None
    calcinfo_nbeta: Optional[int] = None

    # Core properties
    energy: Optional[float] = None
    gradient: Optional[ArrayLike2D] = None
    hessian: Optional[ArrayLike2D] = None
    nuclear_repulsion_energy: Optional[float] = None

    wfn: Optional[Wavefunction] = None

    @validator("gradient")
    def validate_gradient_shape(cls, v: ArrayLike2D):
        """Validate gradient is n x 3"""
        return np.asarray(v).reshape(-1, 3)

    @validator("hessian")
    def validate_hessian_shape(cls, v: ArrayLike2D):
        """Validate hessian is square"""
        v = np.asarray(v)
        n = int(np.sqrt(v.size))
        return v.reshape((n, n))


class SinglePointOutputMixin(BaseModel):
    """Output for a single point calculation.

    Attributes:
        input_data: The SinglePointInput object for the computation.
    """

    input_data: "SinglePointInput"


class SinglePointFailedOutput(SinglePointOutputMixin, FailedOutputBase):
    """A failed SinglePoint calculation.

    Attributes:
        success: Always False for a Failed output.
        traceback: String representation of the traceback of the exception that caused
            the failure.
        computed: Any computed data that was able to be extracted before the exception.
        input_data: The input object for the computation.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    input_data: "SinglePointInput"


class SinglePointInput(StructuredInputBase):
    """Input for a single point calculation.

    Attributes:
        program_args: Input arguments for the program.
        molecule: The molecule to be used in the calculation.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    program_args: SinglePointProgramArgs
    molecule: Molecule

    def __repr_args__(self) -> "ReprArgs":
        return [
            ("molecule", self.molecule),
            ("program_args", self.program_args),
        ]

    def get_failed_output_class(self) -> Type[SinglePointFailedOutput]:
        """Return the FailedOutput class for this input class."""
        return SinglePointFailedOutput

    def get_successful_output_class(self) -> Type[SuccessfulOutputBase]:
        return SinglePointSuccessfulOutput


class SinglePointSuccessfulOutput(SinglePointOutputMixin, SuccessfulOutputBase):
    """Result from a successful single point calculation.

    Attributes:
        input_data: The SinglePointInput object for the computation.
        success: Always True for a successful computation.
        computed: The results computed by the program.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the results.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    computed: SinglePointComputedProperties

    def __repr_args__(self) -> "ReprArgs":
        """A helper for __repr__ that returns a list of tuples of the form
        (name, value).
        """
        return [
            ("success", self.success),
            ("input_data", self.input_data),
            ("computed", "(...)"),
            ("provenance", self.provenance),
        ]

    @property
    def return_result(self) -> Union[float, ArrayLike2D]:
        """Return the result of the calculation.

        Returns:
            The explicitly requested result of the calculation, i.e., the energy,
                gradient, or hessian.
        """
        return getattr(self.computed, self.input_data.program_args.calc_type.value)



from pathlib import Path
from typing import List, Union

import numpy as np

StrOrPath = Union[str, Path]

# May be energy (float), gradient or hessian (List[List[float]])
SPReturnResult = Union[float, List[List[float]]]

# Type for any values that can be coerced to 2D numpy array
ArrayLike2D = Union[List[List[float]], List[float], np.ndarray]
